/* sample style sheet for wiliki */

body {
  font-family: verdana, arial, helvetica, sans-serif;
  color: black;
  background-color: #ffffee;
}

:link    { text-decoration: none; color: #00bb00; }
:visited { text-decoration: none; color: #557722; }

h1 { 
  text-align: right;
  font-size: 190%;
  color: #444400;
  margin-top: 20pt;
  margin-bottom: 20pt;
  border-bottom: solid thick #888844;
}

h2 {
  text-align: left;
  font-size: 150%;
  color: #448800;
  margin-top: 30pt;
  border-bottom: solid thin #448800;
}

h3, h4, h5, h6 {
   text-align: left;
   color: #448800;
   background: transparent;
}
h3 { font-size: 130% }
h4 { font-size: 110% }
h5 { font-size: 100% }
h6 { font-size: 100%; font-style: italic }

pre {
    margin-left: 2em;
    margin-right: 2em;
    font-family: monospace;
    background-color: #eeffee;
    border: solid thin #9999aa;
}

pre.macroerror { 
    background-color: #ee5555;
}

blockquote {
    border: solid thick #aabbdd;
}

dt { font-weight: bold }

table.content-table { table-layout : fixed }

td.inbody { background-color: #eeeeee }   /* used in formatted body */
td.preview { background-color: #eeeedd }  /* used in preview page */

div.footer { color: #bbbbbb }

td.menu-strip ul {
  margin: 0pt;
  padding-left: 5pt;
}

div.menu-title {
  text-align : center;
  font-weight: bold;
  background: #aaddbb;
}

input.navi-button {
  margin: 0pt;
  padding: 0pt;
  background-color: #dddddd;
  border: solid thin #777777;
}

input.search-box {
  margin: 0pt;
  padding: 0pt;
  border: solid 1pt #777777;
}

input.navi-button:hover {
  background: #ffffee;
}

span.breadcrumb-links {
  color: #444444;
  font-size: 90%;
}

span.wiliki-alert { background-color: #ff8080; } /* used to warn internal error */

